import sympy as sy
import numpy as np

from .dhmatrix import DHMatrix


class SpeedKinematics:
    @classmethod
    def linearSpeed(cls, transform, args: dict, delta: list):
        yak = cls.jacobiLinear(transform, args)
        return np.dot(yak, np.array(delta).T)

    @staticmethod
    def jacobiLinear(transform, args: dict):
        argNames = list(args.keys())
        return np.array(
            [[sy.diff(transform.row(i).col(3)[0], argNames[j]).subs(args).n() for j in range(len(args))] for i in
             range(3)], dtype='float')

    @classmethod
    def angleSpeed(cls, frames, args: dict, delta: list):
        yak = cls.jacobiAngle(frames, args)
        return np.dot(yak, np.array(delta).T).T

    @staticmethod
    def jacobiAngle(frames, args: dict):
        return np.array([np.array(DHMatrix.rotate(frames[j]).evalf(subs=args).col(2)) for j in range(len(frames) - 1)],
                        dtype='float').T[0]







class Point:

    def __init__(self, coords=None):
        if coords is None:
            coords = [0, 0]
        self.x = coords[0]
        self.y = coords[1]

    def coord(self):
        return [self.x, self.y]


class Triangle:

    def __init__(self, points=None):
        if points is None:
            points = [Point(), Point(), Point()]
        self.point_1 = points[0]
        self.point_2 = points[1]
        self.point_3 = points[2]

    def isIntersection(self, other):
        if self.checkSeparate(self.linePointPairs(), other.points()):
            return False
        if self.checkSeparate(other.linePointPairs(), self.points()):
            return False
        return True

    @classmethod
    def checkSeparate(cls, pairs, points):
        for s_p in pairs:
            duration = cls.isClockWise(s_p[0], s_p[1], s_p[2])
            separate = True
            for p in points:
                if duration == cls.isClockWise(s_p[0], s_p[1], p) or cls.isLine(s_p[0], s_p[1], p):
                    separate = False
                    break
            if separate:
                return True
        return False

    def points(self):
        return [self.point_1, self.point_2, self.point_3]

    def edges(self):
        return [[self.point_1, self.point_2], [self.point_2, self.point_3], [self.point_3, self.point_1]]

    def linePointPairs(self):
        return [[self.point_1, self.point_2, self.point_3],
                [self.point_2, self.point_3, self.point_1],
                [self.point_3, self.point_1, self.point_2]]

    @staticmethod
    def durationRotate(first: Point, second: Point, third: Point):
        return (second.x - first.x) * (third.y - first.y) - (second.y - first.y) * (third.x - first.x)

    @classmethod
    def isClockWise(cls, first: Point, second: Point, third: Point):
        return cls.durationRotate(first, second, third) < 0

    @classmethod
    def isLine(cls, first: Point, second: Point, third: Point):
        return cls.durationRotate(first, second, third) == 0



import sympy as sy


class DHMatrix:
    """Класс для работы с матрицей Денавита-Хартенберга"""

    @staticmethod
    def matrix(r, a, d, theta):
        """Возвращает матрицу Денавита-Хартенберга из входных параметров"""
        return sy.Matrix([[sy.cos(theta), -1 * sy.sin(theta) * sy.cos(a), sy.sin(theta) * sy.sin(a), r * sy.cos(theta)],
                         [sy.sin(theta), sy.cos(theta) * sy.cos(a), -1 * sy.cos(theta) * sy.sin(a), r * sy.sin(theta)],
                         [0, sy.sin(a), sy.cos(a), d],
                         [0, 0, 0, 1]])

    @staticmethod
    def pos(matrix: sy.Matrix) -> sy.Matrix:
        """Возвращает матрицу положения x, y, z из матрицы Денавита-Хартенберга"""
        return sy.Matrix([matrix.row(j).col(3)[0] for j in range(3)])

    @staticmethod
    def rotate(matrix: sy.Matrix) -> sy.Matrix:
        """Возвращает матрицу поворота из матрицы Денавита-Хартенберга"""
        return sy.Matrix([[matrix.row(j).col(i)[0] for i in range(3)] for j in range(3)])



import numpy
import math


class Quaternion:
    def __init__(self, scalar: float = 1.0, vector=None):
        if vector is None:
            vector = [1, 1, 1]
        self._scalar = round(scalar, 3)
        self._vector = numpy.array(vector)

    @classmethod
    def fromAngleAndVector(cls, angle: float = 0.0, v=None):
        if v is None:
            v = [1, 1, 1]
        _scalar = math.cos(angle/2)
        _vector = numpy.array(v) * math.sin(angle/2)
        return cls(_scalar, _vector)

    def __mul__(self, other):
        scalar = self._scalar * other.scalar - numpy.dot(self.vector.transpose(), other.vector)
        vector = self._scalar * other.vector + other.scalar * self.vector + numpy.cross(self.vector, other.vector)
        return Quaternion(scalar, vector)

    def __add__(self, other):
        scalar = self._scalar + other.scalar
        vector = self._vector + other.vector
        return Quaternion(scalar, vector)

    def __sub__(self, other):
        scalar = self._scalar - other.scalar
        vector = self._vector - other.vector
        return Quaternion(scalar, vector)

    def scalarMulti(self, other) -> float:
        return (other.scalar * self.scalar + other.x * self.x + other.y * self.y + other.z * self.z) \
               / (self.norma() * other.norma())

    def multiToScalar(self, s):
        scalar = self._scalar *  s
        vector = self._vector * s
        return Quaternion(scalar, vector)

    def toRotateMatrix(self):
        a11 = 1 - 2 * self.y ** 2 - 2 * self.z ** 2
        a12 = 2 * self.x * self.y - 2 * self.z * self.scalar
        a13 = 2 * self.x * self.z + 2 * self.y * self.scalar

        a21 = 2 * self.x * self.y + 2 * self.z * self.scalar
        a22 = 1 - 2 * self.x ** 2 - 2 * self.z ** 2
        a23 = 2 * self.y * self.z - 2 * self.x * self.scalar

        a31 = 2 * self.x * self.z - 2 * self.y * self.scalar
        a32 = 2 * self.y * self.z + 2 * self.x * self.scalar
        a33 = 1 - 2 * self.x ** 2 - 2 * self.y ** 2

        R = numpy.array([[a11, a12, a13],
                         [a21, a22, a23],
                         [a31, a32, a33]])
        return R

    def norma(self):
        return math.sqrt(self._scalar ** 2 + numpy.dot(self._vector.transpose(), self._vector))

    def isUnit(self):
        return round(self.norma()) == 1

    @property
    def scalar(self):
        return self._scalar

    @scalar.setter
    def scalar(self, value):
        self._scalar = value

    @property
    def vector(self):
        return self._vector

    @vector.setter
    def vector(self, value):
        self._vector = value

    @property
    def x(self):
        return self._vector[0]

    @x.setter
    def x(self, value):
        self._vector[0] = value

    @property
    def y(self):
        return self._vector[1]

    @y.setter
    def y(self, value):
        self._vector[1] = value

    @property
    def z(self):
        return self._vector[2]

    @z.setter
    def z(self, value):
        self._vector[2] = value



import math
import numpy as np


def rotateMatrixToAxisAndAngle(r):
    theta = math.acos((np.trace(r) - 1) / 2)
    if round(theta, 5) == 0:
        return theta, np.array([None, None, None]), None
    if math.pi - 0.001 <= round(theta, 5) <= math.pi + 0.001:
        wx = math.sqrt((r[0][0] + 1) / 2)
        wy = math.sqrt((r[1][1] + 1) / 2) if wx == 0 else r[0][1] / (2 * wx)
        if wx != 0:
            wz = r[0][2] / (2 * wx)
        elif wy != 0:
            wz = r[1][2] / (2 * wy)
        else:
            wz = math.sqrt((r[2][2] + 1) / 2)
        w = np.array([wx, wy, wz])
        return theta, w, -1 * w
    w = np.array([r[2][1] - r[1][2],
                  r[0][2] - r[2][0],
                  r[1][0] - r[0][1]])
    w = w/(2 * math.sin(theta))
    return theta, w, None




import sympy as sy

from .dhmatrix import DHMatrix


class ForwardKinematics:
    """Класс для работы с прямой кинематикой"""

    @staticmethod
    def links(d: list, r: list, thetas: list, alpha: list) -> list:
        """Формирование всех матриц преобразования T(i-1, i)"""
        return [DHMatrix.matrix(r[i], alpha[i], d[i], thetas[i]) for i in range(len(d))]

    @staticmethod
    def frames(links: list) -> list:
        """Формирование всех матриц преобразования T(0, i) где i от 0 до N"""
        frames = list([sy.eye(4)])
        frames.append(links[0])
        for i in range(1, len(links)):
            frames.append(frames[-1] * links[i])
        return frames

    @staticmethod
    def resultMatrix(links: list) -> sy.Matrix:
        """Формирование результирующей матрицы преобразования"""
        fk = links[0]
        for i in range(1, len(links)):
            fk = fk * links[i]
        return fk



import math

from .quaternion import Quaternion


def nextQ(q0: Quaternion, q1: Quaternion, theta: float, t: int):
    firstPart = math.sin((1 - t) * theta) / math.sin(theta)
    secondPart = math.sin(t * theta) / math.sin(theta)
    return q0.multiToScalar(firstPart) + q1.multiToScalar(secondPart)


def quatSlerp(q0: Quaternion, q1: Quaternion, step: int = 3):
    cosTheta = q0.scalarMulti(q1)
    theta = math.acos(cosTheta)
    if theta > math.pi:
        theta -= 2 * math.pi
    quaternions = list()
    tList = [i / step for i in range(step + 1)]
    for t in tList:
        quaternions.append(nextQ(q0, q1, theta, t))
    return quaternions



import math

import numpy as np


def GradientBasedPlanner(f, start_coords, end_coords, max_its):
    gx = np.gradient(f, axis=0)
    gy = np.gradient(f, axis=1)
    route = [start_coords]
    cur_coords = np.array(start_coords)
    for i in range(max_its):
        direction = np.array([gx[int(cur_coords[0])][int(cur_coords[1])],
                              gy[int(cur_coords[0])][int(cur_coords[1])]])
        direction = direction / math.sqrt(direction[0] ** 2 + direction[1] ** 2)

        cur_coords = np.array([cur_coords[0] - direction[0], cur_coords[1] - direction[1]])

        if cur_coords[0] < 1:
            cur_coords[0] = 0
        elif cur_coords[0] > len(f) - 1:
            cur_coords[0] = len(f) - 1
        if cur_coords[1] < 1:
            cur_coords[1] = 0
        elif cur_coords[1] > len(f[0]) - 1:
            cur_coords[1] = len(f[0]) - 1

        route.append(list(map(int, cur_coords)))
        dist_to_end = end_coords - cur_coords
        if dist_to_end[0] ** 2 + dist_to_end[1] ** 2 < 4:
            route.append(end_coords)
            break
    return route



import random
from math import pi

import numpy as np


def xRotation(vector, theta):
    """Поворот 3D вектора вокруг оси x"""
    R = np.array([[1, 0, 0],
                  [0, np.cos(theta), -np.sin(theta)],
                  [0, np.sin(theta), np.cos(theta)]])
    return np.dot(R, vector)


def yRotate(vector, theta):
    """Поворот 3D вектора вокруг оси y"""
    R = np.array([[np.cos(theta), 0, np.sin(theta)],
                  [0, 1, 0],
                  [-np.sin(theta), 0, np.cos(theta)]])
    return np.dot(R, vector)


def zRotate(vector, theta):
    """Поворот 3D вектора вокруг оси z"""
    R = np.array([[np.cos(theta), -np.sin(theta), 0],
                  [np.sin(theta), np.cos(theta), 0],
                  [0, 0, 1]])
    return np.dot(R, vector)


def randomRotate(vector):
    """Поворот 3D вектора на рандомные углы вокруг оси x, y и z"""
    x = random.randint(int(-pi * 100), int(pi * 100)) / 100
    y = random.randint(int(-pi * 100), int(pi * 100)) / 100
    z = random.randint(int(-pi * 100), int(pi * 100)) / 100
    return zRotate(yRotate(xRotation(vector, x), y), z)



from .routecommon import *


class NodeDij(Node):
    distance = float('inf')


def minDistId(list_id: list, graph: dict):
    min_dist_id = list_id[0]
    for i in list_id:
        if graph[i].distance < graph[min_dist_id].distance:
            min_dist_id = i
    return min_dist_id


def dijkstraGrid(input_map, start_coords, dest_coords):
    graph = createGraph(input_map, NodeDij)
    start_id = idByCoord(start_coords, graph)
    graph[start_id].distance = 0
    not_watched = list([start_id])
    while len(not_watched) != 0:
        current_id = minDistId(not_watched, graph)
        not_watched.remove(current_id)
        graph[current_id].is_watched = True
        nearId = neighbors(current_id, graph)
        for n_id in nearId:
            if graph[n_id].distance > graph[current_id].distance + 1:
                graph[n_id].distance = graph[current_id].distance + 1
                graph[n_id].parent_id = current_id
            if not graph[n_id].is_watched:
                not_watched.append(n_id)
    return pathFromTo(graph, start_coords, dest_coords)



import copy


class Node:
    parent_id = -1
    coord = [-1, -1]
    is_watched = False


def createGraph(input_map: list, node_class):
    graph = dict()
    count = 0
    for x in range(len(input_map)):
        for y in range(len(input_map[x])):
            if not input_map[x][y]:
                n = node_class()
                n.coord = [x, y]
                graph[count] = n
                count += 1
    return graph


def idByCoord(coord: list, graph: dict):
    for id in graph.keys():
        if graph[id].coord == coord:
            return id
    return -1


def neighbors(id: list, graph: dict):
    coord = graph[id].coord

    nearCoords = [[coord[0] + 1, coord[1]],
            [coord[0] - 1, coord[1]],
            [coord[0], coord[1] + 1],
            [coord[0], coord[1] - 1]]

    result = list()
    for n in nearCoords:
        id_n = idByCoord(n, graph)
        if id_n != -1:
            result.append(id_n)
    return result


def pathFromTo(graph: dict, start_coords: list, dest_coords: list):
    path = list([dest_coords])
    current_coord = dest_coords
    dest_parent = graph[idByCoord(current_coord, graph)].parent_id
    if dest_parent == -1:
        return list()
    while current_coord != start_coords:
        parent_id = graph[idByCoord(current_coord, graph)].parent_id
        current_coord = graph[parent_id].coord
        path.append(current_coord)
    return path[::-1]


def startAndEntToMap(input_map, start_coords, dest_coords):
    map_with_start_end = copy.deepcopy(input_map)
    map_with_start_end[start_coords[0]][start_coords[1]] = 5
    map_with_start_end[dest_coords[0]][dest_coords[1]] = 6
    return map_with_start_end


def routeToMap(input_map, route, color=7):
    map_with_all = copy.deepcopy(input_map)
    for r in range(1, len(route) - 1):
        input_map[route[r][0]][route[r][1]] = color
    return map_with_all


import math

from .routecommon import *


class NodeA(Node):
    g = float('inf')
    f = float('inf')


def H(node_coord, dest_coords):
    return math.sqrt((node_coord[0] - dest_coords[0])**2 + (node_coord[1] - dest_coords[1])**2)


def minFId(list_id, graph):
    min_f_id = list_id[0]
    for i in list_id:
        if graph[i].f < graph[min_f_id].f:
            min_f_id = i
    return min_f_id


def AStarGrid(input_map, start_coords, dest_coords):
    graph = createGraph(input_map, NodeA)
    start_id = idByCoord(start_coords, graph)
    graph[start_id].g = 0
    graph[start_id].f = H(graph[start_id].coord, dest_coords)
    not_watched = list([start_id])
    while len(not_watched) != 0:
        current_id = minFId(not_watched, graph)
        not_watched.remove(current_id)
        graph[current_id].is_watched = True
        nearId = neighbors(current_id, graph)
        for n_id in nearId:
            if graph[n_id].g > graph[current_id].g + 1:
                graph[n_id].g = graph[current_id].g + 1
                graph[n_id].f = graph[current_id].f + H(graph[n_id].coord, dest_coords)
                graph[n_id].parent_id = current_id
            if not graph[n_id].is_watched:
                not_watched.append(n_id)
    return pathFromTo(graph, start_coords, dest_coords)
